#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2023 "The Tesseract" November 20, 2023

in vec3 in_Position;
out vec4 uv;
out float time;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};
layout(std140) uniform MidiBlock {
	vec4 midi[64];//x = note id, y = velocity, z = channel/track, w = time in seconds since note pressed
	int midiCount;
};
/*LMMS midi converter https://www.lynxwave.com/LMMStoMIDI/LMMStoMIDIConverter.html since LMMS midi export is broken.

Launch options for midi config 'Shaderview console art/TheTesseract.mid 15 0.5'
15 is the bitmask 0b001111 signifying which instruments are beats/drums with a fixed length.
0.5 is the fixed beat length in seconds.

Midi Instrument 0 = ACCENTHH.wav
Midi Instrument 1 = snare03.ogg
Midi Instrument 2 = RY30Kick03.wav
Midi Instrument 3 = RY30Kick10.wav
Midi Instrument 4 = glitch
Midi Instrument 5 = Fathom(AwwKEYline)*/

//x = rot ang 1, y = rot ang 2, z = kick, w = snare/hihat
out vec4 music;
out mat2 tsRot[3];



#define AA_SAMPLES 4
#define ANIMATION 1

#define PRECISION .035

#define PI 3.141592653589793


mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}
//r2 hash http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
vec2 quasiRandom(float n) {
	#define R2V 1.32471795724474602596
	return fract(.5+n*(1./vec2(R2V,R2V*R2V)));
}



vec4 Volume(vec3 p, vec4 col) {
	//3d -> 4d projection
	vec4 tp = vec4(0,0,0,-2),
		td = normalize(vec4(p,1));
	
	//rotate
	tp.xw *= tsRot[0];
	td.xw *= tsRot[0];
	
	tp.yw *= tsRot[1];
	td.yw *= tsRot[1];
	
	tp.yz *= tsRot[2];
	td.yz *= tsRot[2];
	
	
	//volume ray march
	for (int i = 0; i < 32 && length(tp) < 2.1 && col.w < 1.; i++) {
		//4d line box sdf
		vec4 dp = abs(tp)-.9, bp = abs(dp)-.1;
		float bounds = max(max(dp.x,dp.y),max(dp.z,dp.w)),
			dst = min( min(max(bp.x,max(bp.y,bp.z)), max(bp.y,max(bp.z,bp.w))),
					min(max(bp.z,max(bp.w,bp.x)), max(bp.w,max(bp.x,bp.y))));
		dst = max(dst,bounds);
		
		if (dst < .3) {
			//solid gray lines
			float alpha = clamp(1.-dst/.01,0.,1.);
			if (alpha > 0.) {
				col += vec4(vec3((.5+bounds*5.)*(.035+.15*fract(length(music)))), 1)*(alpha*100.*PRECISION);
			}
		
			//volumetric light
			float valpha = clamp(1.-dst/.3,0.,1.);
			if (valpha > 0.) {
				col += vec4(music.z*(.5+tp.y*.4), 0, music.w*max(0., .5+tp.w)*3., 1)*
						(valpha/max(1., dst*dst*500.)*max(.01,dst)*4.*PRECISION);
			}
		}

		tp += td*dst;
	}
	
	return col;
}



uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION
	time = float(frame)/float(framerate*AA_SAMPLES);

	music.xy = texelFetch(Buffer0,ivec2(0),0).xy;
	if (frame == 0) music.xy = vec2(0);
	
	float delta = 1./float(framerate*AA_SAMPLES),
		gscale = min(1.,time/10.);
	music.xy = max(music.xy-vec2(.5,2)*delta, 0.);
	music.zw = vec2(0);
	for (int i = 0; i < midiCount; i++) {
		switch (int(midi[i].z)) {
			case 0://hihat
				music.w += max(0.,1.-midi[i].w/.1)*.5;
				break;
		
			case 1://snare
				music.w += max(0.,1.-midi[i].w/.25);
				break;
				
			case 2://kick
				music.z += max(0.,1.-midi[i].w/.2);
				break;
				
			case 3://crash kick
				music.z += max(0.,1.-midi[i].w/.3);
				break;
			
			case 4://glitch
				music.x += delta*2.*gscale;
				break;
				
			case 5://bass
				music.y += delta*3.;
				break;
		}
	}
	
	#else
	//screenshot 1
	time = 0.;
	music = vec4(1., 3.7, 1,1);
	
	//screenshot 2
	//time = 32.;
	//music = vec4(1.2,1.3,1,1);
	#endif
	
	//precalculate tesseract rotations
	tsRot[0] = r2d(music.y);
	tsRot[1] = r2d(music.x);
	tsRot[2] = r2d(music.y*.37892);
	
	uv = vec4((in_Position.xy*vec2(screenX/screenY,1)+(quasiRandom(float(frame%AA_SAMPLES))-.5)*2./screenY),
		(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	if (dot(uv.zw,uv.zw) < .8) {
		gl_FragColor = vec4(music.xy,0,1e8);
		return;
	} 

	vec3 rp = vec3(0,0,-2),
		rd = normalize(vec3(uv.xy,1));
		
	//rotate around
	mat2 rm = r2d(time/-30.*PI);
	rp.xz *= rm;
	rd.xz *= rm;
	
	//dither
	float dither = fract(dot(cos(float(frame)/float(AA_SAMPLES)*vec2(3.,1.)+uv.zw), vec2(2373.2397863,2774.7861782)));
	rp += rd*PRECISION*1.*dither;
	
	//volume ray marching
	vec4 c = vec4(0);
	for (int i = 0; i < 100 && c.w < 1.; i++) {
		c = Volume(rp, c);
				
		rp += rd*PRECISION;
	}
	
	if (c.w > 1.) c /= c.w;
	
	
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (frame%AA_SAMPLES == 0) sum = vec4(0);
	sum += vec4(c.xyz, 1.);
	gl_FragColor = sum;
}
